% *************************************************************************
% **************************** The OEOA Toolbox ***************************
% The On-line Ensemble Using Ordered Aggregation (OEOA) Toolbox
% *************************************************************************
% An Adaptive Ensemble of On-line Extreme Learning Machines with Variable 
% Forgetting Factor for Dynamic System Prediction
% *************************************************************************
% Version 1.0 Nov/27/2015
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com 
%
% Homepage: http://sites.google.com/site/symonesoares/
%   
% Paper: SOARES, Symone Gomes; ARAÚJO, Rui. An Adaptive Ensemble of On-line 
% Extreme Learning Machines with Variable Forgetting Factor for Dynamic 
% System Prediction. Neurocomputing, v. 171, p. 693-707, 2016. 
%        DOI: http://dx.doi.org/10.1016/j.neucom.2015.07.035
%
% *************************************************************************
% *************************************************************************
% An Example of the *On-line Sequential Extreme Learning Machine (OS-ELM) 
% algorithm and the **On-line Sequential Extreme Learning Machine using 
% with Directional Forgetting Factor (DFF-OS-ELM) algorithm using the
% ***Debutanizer data set.
%
%   *LIANG, N.-Y.; HUANG, G.-B.; SARATCHANDRAN, P.; and SUNDARARAJAN, N. 
%    (2006). A Fast and Accurate Online Sequential Learning Algorithm for 
%    Feedforward Networks. IEEE Transactions on Neural Networks, 17(6), 
%    1411–1423.
%
%
%  **SOARES, Symone Gomes; ARAÚJO, Rui. An Adaptive Ensemble of On-line 
%    Extreme Learning Machines with Variable Forgetting Factor for Dynamic 
%    System Prediction. Neurocomputing, v. 171, p. 693-707, 2016. 
%
%
% ***FORTUNA, L.; GRANIANI, S.; RIZZO, A.; XIBILIA, M. G. Soft Sensors for 
%    Monitoring and Control of Industrial Processes (Advances in Industrial
%    Control), Springer-Verlag New York, Inc., Secaucus, NJ, USA, 2006.
%
% *************************************************************************

clear; clc; close all;

% Loading the Debutanizer data set
D      = load('Debutanizer.txt');
input  = D(:,1:7); 
output = D(:,8);

% Parameter Setting
fprintf('*** Testing The OEOA Toolbox***\n');
fprintf('    Example of the OS-ELM algorithm and the DFF-OS-ELM algorithm.\n\n');
kfold   = 10;    % number of folds for the k-fold cross-validation
sW      = 50;    % window's size (sW >= kfold)
maxNHL  = 20;    % maximum number of hidden nodes in OS-ELM (maxHL >= sW)
ActF    = 'sig'; % activation function for OS-ELM   

% Other setting
XShifted     = input(1:sW,:);  %Assigning the inputs for the initial model
YShifted     = output(1:sW,:); %Assigning the outputs for the initial model
[Tmse, ~, ~] = KfoldCV_OSELM_ZeroMeanUnitVar(XShifted, YShifted,  kfold, 1, [1 maxNHL], ActF);
[~, nHL]     = min(Tmse); 
nS = size(input,1);

% Calling fOEOA_OSELM with no ordered aggregation
fprintf('*** Starting the initial training of the OS-ELM algorithm ***\n\n');
[mxcal1, stdx1, mycal1, stdy1, mcxcal1, mcycal1] = scaleZmean1Var(XShifted,YShifted);
varx1 = (stdx1).^2;
vary1 = (stdy1) ^2;
[~, ~, ELM1] = OSELM_Reg_Train(mcxcal1, mcycal1, nHL, ActF);

% Calling fOEOA_OSELM with no ordered aggregation
fprintf('*** Starting the initial training of the DFF-OS-ELM algorithm ***\n\n');
[mxcal2, stdx2, mycal2, stdy2, mcxcal2, mcycal2] = scaleZmean1Var(XShifted,YShifted);
varx2 = (stdx2).^2;
vary2 = (stdy2) ^2;
[~, ~, ELM2] = fDFF_OSELM_Train(mcxcal2, mcycal2, nHL, ActF);

% Retraining of OS-ELM and DFF-OS-ELM
fprintf('*** Retraining of OS-ELM and DFF-OS-ELM ***\n');

YP1 = []; YP2 = []; YR = []; S1 = sW; S2 = sW;
for i = sW+1:nS

   % Receiving a new sample
   xi = input(i,:);
   yi = output(i,:);

   % Predicting xi using OS-ELM 
   scy1         = scale(xi, mxcal1, stdx1);            %center test X-block
   [~, yp_aux1] = OSELM_Reg_Test(scy1, zeros(1,size(xi,1)), ELM1);
   yp1          = rescale(yp_aux1, mycal1, stdy1);
   
   % Predicting xi using OS-ELM 
   scy2         = scale(xi, mxcal2, stdx2);            %center test X-block
   [~, yp_aux2] = OSELM_Reg_Test(scy2, zeros(1,size(xi,1)), ELM2);
   yp2          = rescale(yp_aux2, mycal2, stdy2);
 
   % Backuping predicted outputs
   YP1 = [YP1 yp1];
   YP2 = [YP2 yp2];
   YR  = [YR yi];
   
   % The OS-ELM retraining
   oldxmean = mxcal1; 
   oldymean = mycal1;
   mxcal1   = S1 / (S1 + 1) * mxcal1 + (1 / (S1 + 1)) * xi;
   mycal1   = S1 / (S1 + 1) * mycal1 + (1 / (S1 + 1)) * yi;     
   varx1    = ((S1 - 1)/ (S1)) * varx1 + (1 / S1) * ((xi - mxcal1).^2) + (mxcal1 - oldxmean).^2;
   vary1    = ((S1 - 1)/ (S1)) * vary1 + (1 / S1) * ((yi - mycal1).^2) + (mycal1 - oldymean).^2;
   stdx1    = sqrt(varx1);
   stdy1    = sqrt(vary1);
   S1       = S1 + 1;
   xiS            = scale(xi, mxcal1, stdx1);
   yiS            = scale(yi, mycal1, stdy1);
   [ELM1, ~] = OSELM_Reg_Retrain(ELM1, xiS, yiS);
   
   % The DFF-OS-ELM retraining
   oldxmean = mxcal2; 
   oldymean = mycal2;
   mxcal2   = S2 / (S2 + 1) * mxcal2 + (1 / (S2 + 1)) * xi;
   mycal2   = S2 / (S2 + 1) * mycal2 + (1 / (S2 + 1)) * yi;     
   varx2    = ((S2 - 1)/ (S2)) * varx2 + (1 / S2) * ((xi - mxcal2).^2) + (mxcal2 - oldxmean).^2;
   vary2    = ((S2 - 1)/ (S2)) * vary2 + (1 / S2) * ((yi - mycal2).^2) + (mycal2 - oldymean).^2;
   stdx2    = sqrt(varx2);
   stdy2    = sqrt(vary2);
   S2       = S2 + 1;
   xiS            = scale(xi, mxcal2, stdx2);
   yiS            = scale(yi, mycal2, stdy2);
   [ELM2, ~] = fDFF_OSELM_Retrain(ELM2, xiS, yiS);
   
end

% Printing results
fprintf('\n*** RESULTS ***');
fprintf('\n\nOS-ELM - Correlation coefficient: %f \n', corr2(YR,YP1));
fprintf('OS-ELM - Mean Squared Error:      %f \n', mse(YP1 - YR));
fprintf('OS-ELM - Root Mean Squared Error: %f \n', mse(YP1 - YR)^1/2);

fprintf('\n\nDFF-OS-ELM - Correlation coefficient: %f \n', corr2(YR,YP2));
fprintf('DFF-OS-ELM - Mean Squared Error:      %f \n', mse(YP2 - YR));
fprintf('DFF-OS-ELM - Root Mean Squared Error: %f \n', mse(YP2 - YR)^1/2);

% Plotting Results
fprintf('\n*** Ploting Testing Data ***\n');
figure;
plot(YR, 'k');
hold on;
plot(YP1, 'r');
hold on;
plot(YP2, 'b');
xlim([1 size(YR,2)]);
legend('Real Output', 'Predicted Output - OS-ELM', 'Predicted Output - DFF-OS-ELM', 2);
