% *************************************************************************
% **************************** The OEOA Toolbox ***************************
% The On-line Ensemble Using Ordered Aggregation (OEOA) Toolbox
% *************************************************************************
% An Adaptive Ensemble of On-line Extreme Learning Machines with Variable 
% Forgetting Factor for Dynamic System Prediction
% *************************************************************************
% Version 1.0 Nov/27/2015
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com 
%
% Homepage: http://sites.google.com/site/symonesoares/
%   
% Paper: SOARES, Symone Gomes; ARAÚJO, Rui. An Adaptive Ensemble of On-line 
% Extreme Learning Machines with Variable Forgetting Factor for Dynamic 
% System Prediction. Neurocomputing, v. 171, p. 693-707, 2016. 
%        DOI: http://dx.doi.org/10.1016/j.neucom.2015.07.035
%
% *************************************************************************
% *************************************************************************
% *************************************************************************
% Algorithm Description:
%    The retraining of the DFF-OS-ELM algorithm for regression 
% *************************************************************************
% Inputs of the Algorithm:
%
% - ELM: parameters of the DFF-OS-ELM model
%
% - xi(1 x number of input variables) : input vector
% 
% - yi(output variable): output point
%   ps.: it accepts only one output variable
%
% *************************************************************************
% Outputs of the Algorithm:
%
% - ELM: new parameters of the DFF-OS-ELM model
%
% - yiP: predicted output after retraining the DFF-OS-ELM model
%
% *************************************************************************
% *************************************************************************
% *************************************************************************


function [ELM, yiP] = fDFF_OSELM_Retrain(ELM, xi, yi)
    
    %%%%%%%%%%%%% step 2 Sequential Learning Phase
    switch lower(ELM.ActivationFunction)
        case{'rbf'}
            H = RBFun(xi,ELM.IW,ELM.Bias);
        case{'sig'}
            H = SigActFun(xi,ELM.IW,ELM.Bias);
        case{'sin'}
            H = SinActFun(xi,ELM.IW,ELM.Bias);
        case{'hardlim'}
            H = HardlimActFun(xi,ELM.IW,ELM.Bias);
    end
    
    %Update Beta
    ep       = yi - H * ELM.beta; 
    ks       = H * ELM.M * H';
    pp       = ((ELM.M * H')/(1 + ks)) * ep; 
    ELM.beta = ELM.beta + pp; 
    
    %Updating other parameters
    if ks > 0
        eps   = ELM.l - (1 - ELM.l)/ks; 
        ELM.M = ELM.M - (ELM.M * (H') * H * ELM.M)/(inv(eps) + ks); 
    end
    
    ELM.la = ELM.l * (ELM.la + (ep * ep)/(1 + ks));
    ELM.ny = ELM.l * (ELM.ny + 1); 
    te     = (ep * ep)/ELM.la;
    ELM.l  =  1/(1+(1+ELM.ro) * (log(1+ks) + ((((ELM.ny+1)*te)/(1+ks+te))- 1) * (ks/(1+ks))));
    
    yiP    = H * ELM.beta;

end
 