% *************************************************************************
% **************************** The OEOA Toolbox ***************************
% The On-line Ensemble Using Ordered Aggregation (OEOA) Toolbox
% *************************************************************************
% An Adaptive Ensemble of On-line Extreme Learning Machines with Variable 
% Forgetting Factor for Dynamic System Prediction
% *************************************************************************
% Version 1.0 Nov/27/2015
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com 
%
% Homepage: http://sites.google.com/site/symonesoares/
%   
% Paper: SOARES, Symone Gomes; ARAÚJO, Rui. An Adaptive Ensemble of On-line 
% Extreme Learning Machines with Variable Forgetting Factor for Dynamic 
% System Prediction. Neurocomputing, v. 171, p. 693-707, 2016. 
%        DOI: http://dx.doi.org/10.1016/j.neucom.2015.07.035
%
% *************************************************************************
% *************************************************************************
% *************************************************************************
% Algorithm Description:
%    The OEOA algorithm using the On-line Sequential Extreme
%    Learning Machine with Direction Forgetting Factor (DFF-OS-ELM)
% *************************************************************************
% Inputs of the Algorithm:
%
% - input(number of samples x number of input variables) : input matrix
% 
% - output(number of samples x output variable) : output matrix 
%   ps.: it accepts only one output variable
%
% - kfold   : number of folds for the k-fold cross-validation
%             (kfold < number of samples) 
%             Default: 10; 
% - sW      : window's size (sW >= kfold) and (sW < number of samples) 
% - maxNHL  : maximum number of hidden nodes in DFF-OS-ELM (maxNHL >= sW) 
%             Default: 20; (and maxNHL = sW if maxNHL > sW) 
% - Alpha   : factor to include a new model (0 < Alpha < 1);
%             Default: 0.04; 
% - ActF    : activation function for OS-ELM
%             Options: 'rbf'; 'sig'; 'sin'; 'hardlim';
%             Default: 'sig';
% - M_max   : maximum number of models in the ensemble (M_max > 1) and 
%             (M_max >= M_min)
%             Default: 30;
% - M_min   : minimum number of models in the ensemble (M_min > 1) and
%             (M_max >= M_min)
%             Default: 10;
%  
%  ps.: If M_max = M_min, then no Ordered Aggregation is applied
% 
% *************************************************************************
% Outputs of the Algorithm:
%
% CC   : correlation coefficient between predicted outputs and real outputs
% MSE  : mean square errord between predicted outputs and real outputs
% RMSE : root mean square errord between predicted outputs and real outputs
% YR   : real outputs (testing data, i.e.\ Total of Samples minus sW)
% YP   : predicted outputs (testing data, i.e.\ Total of Samples minus sW)
% time : running time of the algorithm
%
% *************************************************************************
% *************************************************************************
% *************************************************************************

function [CC, MSE, RMSE, YR, YP, time, AllRank] = fOEOA_DFF_OSELM(input, output, sW, maxNHL, Alpha, kfold, ActF, M_max, M_min)
    
    tStart = tic;
    %Setting some parameters 
    nS = size(input,1); YP = []; YR = [];
    if maxNHL > sW
         maxNHL = sW;
    end
    if M_min > M_max
         fprintf('\nError: M_max must be greater than or equal to M_min!!!\n');
         return;
    end   
    
    %Build the initial model
    XShifted              = input(1:sW,:);  %Assigning the inputs for the initial model
    YShifted              = output(1:sW,:); %Assigning the outputs for the initial model
    for nM = 1:M_max
        [Tmse, ~, ~] = KfoldCV_OSELM_ZeroMeanUnitVar(XShifted, YShifted,  kfold, 1, [1 maxNHL], ActF);
        [~, Model(nM).nHL] = min(Tmse); 
        [Model(nM).mxcal,Model(nM).stdx, Model(nM).mycal, Model(nM).stdy, mcxcal, mcycal] = scaleZmean1Var(XShifted,YShifted);
        Model(nM).varx = (Model(nM).stdx).^2 ;
        Model(nM).vary = (Model(nM).stdy) ^2;
        [~, ~, Model(nM).ELM] = fDFF_OSELM_Train(mcxcal, mcycal, Model(nM).nHL, ActF);
        Model(nM).w           = 1/M_max;
        Model(nM).S           = sW;
        Model(nM).life        = 0;
        Model(nM).e           = [];
        Model(nM).mse         = 0;
    end
    RankTh  = M_max;
    AllRank = [];

    %Loop of samples
    for i = sW+1:nS

       xi       = input(i,:);
       yi       = output(i,:);
       yiModels = zeros(1,size(Model,2)); 
       MSEALL   = [];
       for m = 1:size(Model,2)
          scxtest       = scale(xi, Model(m).mxcal, Model(m).stdx);            %center test X-block
          [~, yppred]   = OSELM_Reg_Test(scxtest, zeros(1,size(xi,1)), Model(m).ELM);
          yiModels(:,m) = rescale(yppred, Model(m).mycal, Model(m).stdy);
          MSEALL        = [MSEALL Model(m).mse];
       end

       %Ranking models for prediction
       if M_min ~= M_max
           [~, IXrank]         = sort(MSEALL);
           IXEns               = IXrank(1,1:RankTh);
           IXEns               = sort(IXEns);
           MSEALLaux           = MSEALL(1,IXEns);
       else
           IXEns               = 1:M_max;
           MSEALLaux           = MSEALL;
       end
       wiModelsSelectedAUX = zeros(1, size(IXEns,2));
       for m = 1:size(IXEns,2)
           if Model(IXEns(1,m)).life > 0
               wiModelsSelectedAUX(1,m) = exp(-(Model(IXEns(1,m)).mse - median(MSEALLaux))/(median(MSEALLaux)));
           else
               wiModelsSelectedAUX(1,m) = 1;
           end
       end
       wiModelsSelectedAUX = wiModelsSelectedAUX ./ sum(wiModelsSelectedAUX);
       yiModels2           = yiModels(:,IXEns);    
       yiEns               = wiModelsSelectedAUX * yiModels2';
       YP                  = [YP yiEns];
       YR                  = [YR yi];

       %Sliding the window
       XShifted(1,:) = [];
       YShifted(1,:) = [];
       XShifted      = [XShifted; xi];
       YShifted      = [YShifted; yi];

       %Ranking models for the next generation
       if M_min ~= M_max
           bestV = inf;
           for k = M_min:size(IXrank,2)
               IX2                 = IXrank(1,1:k);
               IX2                 = sort(IX2);
               wiModelsSelectedAUX = zeros(1, size(IX2,2));
               MSEALLaux           = MSEALL(1,IX2);
               for m = 1:size(IX2,2)
                   if Model(IX2(1,m)).life > 0
                       wiModelsSelectedAUX(1,m) = exp(-(Model(IX2(1,m)).mse - median(MSEALLaux))/(median(MSEALLaux)));
                   else
                       wiModelsSelectedAUX(1,m) = 1;
                   end
               end
               wiModelsSelectedAUX = wiModelsSelectedAUX ./ sum(wiModelsSelectedAUX);
               yiModels2           = yiModels(:,IX2);    
               yiEns2              = wiModelsSelectedAUX * yiModels2';
               eNow                = (yi - yiEns2)^2;
               if eNow < bestV
                  RankTh = k; 
                  bestV  = eNow;
               end
           end
       else
           RankTh = M_max; 
       end
       AllRank = [AllRank RankTh];

       %Model retraining
       for m = 1:size(Model,2)
            oldxmean       = Model(m).mxcal; 
            oldymean       = Model(m).mycal;
            Model(m).mxcal = Model(m).S / (Model(m).S + 1) * Model(m).mxcal + (1 / (Model(m).S + 1)) * xi;
            Model(m).mycal = Model(m).S / (Model(m).S + 1) * Model(m).mycal + (1 / (Model(m).S + 1)) * yi;     
            Model(m).varx  = ((Model(m).S - 1)/ (Model(m).S)) * Model(m).varx + (1 / Model(m).S) * ((xi - Model(m).mxcal).^2) + (Model(m).mxcal - oldxmean).^2;
            Model(m).vary  = ((Model(m).S - 1)/ (Model(m).S)) * Model(m).vary + (1 / Model(m).S) * ((yi - Model(m).mycal).^2) + (Model(m).mycal - oldymean).^2;
            Model(m).stdx  = sqrt(Model(m).varx);
            Model(m).stdy  = sqrt(Model(m).vary);
            Model(m).S     = Model(m).S + 1;
            xiS            = scale(xi, Model(m).mxcal, Model(m).stdx);
            yiS            = scale(yi, Model(m).mycal, Model(m).stdy);
            [Model(m).ELM, ~] = fDFF_OSELM_Retrain(Model(m).ELM, xiS, yiS);
       end

       %Updating some parameters
       worstModel = -inf; 
       for k = 1:size(Model,2)       
           Model(k).life  = Model(k).life + 1;
           eNow          = (yi - yiModels(1,k))^2;
           Model(k).e    = [Model(k).e eNow];
           if  Model(k).life <= sW
               Model(k).mse = ((Model(k).life - 1) / Model(k).life) * Model(k).mse + (1 / Model(k).life) * eNow;
           else 
               Model(k).mse = Model(k).mse + eNow/sW - Model(k).e(:,1)/sW;
               Model(k).e(:,1) = [];
           end
           if Model(k).mse > worstModel
               worstModel = Model(k).mse; ix = k;
           end
       end

       %Generating a new model
       if (abs((yiEns - yi)/yi) > Alpha) 
            [Tmse, ~, ~]          = KfoldCV_OSELM_ZeroMeanUnitVar(XShifted, YShifted, kfold, 1, [1 maxNHL], ActF);
            [~, Model(ix).nHL]    = min(Tmse);
            [Model(ix).mxcal,Model(ix).stdx, Model(ix).mycal, Model(ix).stdy, mcxcal, mcycal] = scaleZmean1Var(XShifted, YShifted);
            Model(ix).varx        = (Model(ix).stdx).^2 ;
            Model(ix).vary        = (Model(ix).stdy) ^2;
            [~, ~, Model(ix).ELM] = fDFF_OSELM_Train(mcxcal, mcycal, Model(ix).nHL, ActF);
            Model(ix).S           = sW; 
            Model(ix).e           = [];
            Model(ix).mse         = 0; 
            Model(ix).life        = 0; 
       end

    end

    time = toc(tStart);
    CC   = corr2(YR,YP);
    MSE  = mse(YP - YR);
    RMSE = mse(YP - YR)^1/2;
    
    fprintf('\nCorrelation coefficient: %f \n', CC);
    fprintf('Mean Squared Error:      %f \n', MSE);
    fprintf('Root Mean Squared Error: %f \n', RMSE);
    
end